#include "frameviewer.h"

#include <QtCore>
#include <QBoxLayout>
#include <QGraphicsView>


#include "../globalobjects.h"
#include "../core/functions.h"
#include "Dense"


FrameViewer::FrameViewer(QWidget *parent, Qt::WindowFlags f) :
    QWidget(parent, f),
    pixmapItem()
{
    view = new FrameGraphicsView();
    view->setFrameStyle(QFrame::NoFrame);
    connect(view, SIGNAL(resized()), this, SLOT(resize()));
    connect(view, SIGNAL(doubleClicked(int,int,bool)), this, SLOT(doubleClicked(int,int,bool)));

    scene = new QGraphicsScene(this);
    view->setScene(scene);
    view->setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform);

    QBoxLayout* l = new QBoxLayout(QBoxLayout::LeftToRight);
    if(layout()) delete layout();
    setLayout(l);
    l->setContentsMargins(0,0,0,0);

    l->addWidget(view);

    connectToState();


    setDisabled(true);
}


void FrameViewer::connectToState()
{
    this->disconnect(g_state);

    connect(g_state, SIGNAL(hasChanged(int)), this, SLOT(filterStateChange(int)));
}

void FrameViewer::filterStateChange(int f)
{
    if(f & (DataState::CurrentFrameChange | DataState::ObjectsListChange)) reload();
}


void FrameViewer::reload()
{
    qDebug() << "reload frame viewer";
    if(!g_state->videoLoaded) {
        qDebug() << "no vid set";
        setDisabled(true);
        return;
    }
    setDisabled(false);

    scene->clear();

    QImage* img = new QImage(g_state->vid->getFramePath(g_state->currentFrame));
    pixmapItem = scene->addPixmap(QPixmap::fromImage(*img));
    delete img;


    if(g_state->hasObjects()) {
        foreach(TrackedObject* obj, g_state->objectsList) {
            new PolyAreaGraphicsItem(scene, obj->getCurrentKeyFrame());
        }
    }

    resize();
}

void FrameViewer::doubleClicked(int x, int y, bool shift)
{
    QPointF point = view->mapToScene(x,y);

    // threshold points to image rect
    QRectF imageRect = g_mainWindow->frameViewer->pixmapItem->sceneBoundingRect();
    point.setX(qMin(imageRect.width(), qMax(imageRect.x(), point.x())));
    point.setY(qMin(imageRect.height(), qMax(imageRect.y(), point.y())));

    // if shift is pressed, delete a point, if not add a point
    if(shift) {

        QPolygonF poly = g_state->getCurrentObject()->getCurrentKeyFrame()->polygon;

        if(poly.length() == 1) {
            qDebug() << "object would be delted. TODO!";
            return;
        }
        qreal minDist = std::numeric_limits<qreal>::max();
        qreal minIdx = 0;
        for(int i=0; i<poly.length(); ++i) {
            qreal d = Functions::distance(point, poly[i]);
            if(d < minDist) {
                minDist = d;
                minIdx = i;
            }
        }
        QPolygonF newPoly(poly);
        newPoly.remove(minIdx);
        g_state->getCurrentObject()->getCurrentKeyFrame()->setNewPolygon(newPoly);


    // shift is not pressed, add a point
    } else {

        if(!g_state->hasObjects()) {
            g_state->addObject();
            g_mainWindow->reload();
        }

        QPolygonF poly = g_state->getCurrentObject()->getCurrentKeyFrame()->polygon;

        if(poly.length() < 3) {
            poly.append(point);
            if(poly.length() == 3 && !Functions::isCounterClockwise(poly)) {
                QPolygonF newPoly;
                newPoly << poly[0] << poly[2] << poly[1];
                poly = newPoly;
            }
        } else {
            // compute closest polygon-edge and insert point there
            QPointF v(point);
            qreal minDist = std::numeric_limits<qreal>::max();
            qreal minAssignment;
            for(int i=0; i<poly.length(); i++) {
                QPointF a = poly[i];
                QPointF b = poly[(i + 1) % poly.length()];
                QPointF e = b - a;
                QPointF e_n = 1.0/qSqrt(QPointF::dotProduct(e,e)) * e;
                QPointF av = v - a;
                qreal t = QPointF::dotProduct(av,e_n) / qSqrt(QPointF::dotProduct(e,e));
                QPointF vproj;
                qreal t_line = t;
                if(t_line <= 0) {
                    t_line = 0;
                } else if(t >= 1) {
                    t_line = 1;
                }
                vproj = a + t_line*e;
                qreal dist = qSqrt(QPointF::dotProduct(v - vproj, v - vproj));

                if(dist < minDist) {
                    minDist = dist;
                    minAssignment = i + t_line;
                    if(t <= 0) minAssignment += 0.5;
                    if(t >= 1) minAssignment -= 0.5;
                }
                // special case: point closest to corner (= two edges)
                // solution: take the one with the closest angle
                else if((t_line == 0 || t_line == 1) && dist == minDist) {
                    QPointF otherEndpoint = a + (1-t_line)*e;

                    qreal angle;
                    if(t_line == 0) angle = Functions::degAngleCounterClockwise(v - vproj, otherEndpoint - vproj);
                    else angle = Functions::degAngleCounterClockwise(otherEndpoint - vproj, v - vproj);

                    QPointF conflictPoint;
                    if(t_line == 0) conflictPoint = poly[(i - 1 + poly.length()) % poly.length()];
                    else conflictPoint = poly[(i + 2) % poly.length()];

                    qreal conflictAngle;
                    if(t_line == 0) conflictAngle = Functions::degAngleCounterClockwise(conflictPoint - vproj, otherEndpoint - vproj);
                    else conflictAngle = Functions::degAngleCounterClockwise(otherEndpoint - vproj, conflictPoint - vproj);

                    if(conflictAngle < M_PI) {
                        // angle adjustment not needed for acute angles
                        continue;
                    }

                    qreal halfOtherSideConflictAngle = conflictAngle/2.0;

                    if(angle < halfOtherSideConflictAngle) {
                        minAssignment = i + t_line;
                        if(t <= 0) minAssignment += 0.5;
                        if(t >= 1) minAssignment -= 0.5;
                    }

                }
            }
            poly.insert(qFloor(minAssignment)+1, point);
        }

        g_state->getCurrentObject()->getCurrentKeyFrame()->setNewPolygon(poly);

        // adjust type if polygon changes from empty to non-empty
        KeyFrame::Type type = g_state->getCurrentObject()->getCurrentKeyFrame()->type;
        if(type == KeyFrame::NotSet && !poly.empty()) type = KeyFrame::SetVisible;
        else if(type != KeyFrame::NotSet && poly.empty()) type = KeyFrame::NotSet;
        g_state->getCurrentObject()->getCurrentKeyFrame()->setType(type);
    }
}

void FrameViewer::resize()
{
    if(!pixmapItem) return;
    view->setSceneRect(pixmapItem->sceneBoundingRect());
    view->fitInView(pixmapItem->sceneBoundingRect(), Qt::KeepAspectRatio);
}
