#include "polyareagraphicsitem.h"

#include <QDebug>

#include "../globalobjects.h"

PolyAreaGraphicsItem::PolyAreaGraphicsItem(QGraphicsScene* s, KeyFrame *kf) :
    keyframe(kf),
    move(0.0,0.0),
    polyPoints()
{
    setFlag(QGraphicsItem::ItemIsMovable, true);
    s->addItem(this);

    connect(keyframe, SIGNAL(hasChanged(int)), this, SLOT(filterKeyFrameChange(int)));
    connect(g_state, SIGNAL(hasChanged(int)), this, SLOT(filterStateChange(int)));

    updatePolygon();
    updateStyle();

}

void PolyAreaGraphicsItem::redrawFromPolyPoints()
{
    QPolygonF poly;
    foreach(PolyPointGraphicsItem* polyPoint, polyPoints){
        QPointF newPoint = polyPoint->position + polyPoint->move;
        poly.append(newPoint);
    }
    setPolygon(poly);
}

void PolyAreaGraphicsItem::polyPointsChanged()
{
    QPolygonF poly;
    foreach(PolyPointGraphicsItem* polyPoint, polyPoints){
        QPointF newPoint = polyPoint->position;
        poly.append(newPoint);
    }
    keyframe->setNewPolygon(poly);
    if(keyframe->type == KeyFrame::Automatic) keyframe->setType(KeyFrame::SetVisible);
}

void PolyAreaGraphicsItem::updateStyle()
{
    TrackedObject * obj = keyframe->object;
    if(obj->isCurrent()) setZValue(1);
    else setZValue(0);

    QPen pen;
    pen.setColor(obj->color);
    pen.setWidth(2);
//    if(obj->isCurrent()) {
//        pen.setColor(obj->color);
//        pen.setWidth(2);
//    } else {
//        pen.setColor(Qt::gray);
//        pen.setWidth(2);
//    }
    pen.setCosmetic(true);
    setPen(pen);

    foreach(PolyPointGraphicsItem* polyPoint, polyPoints) {
        polyPoint->setVisible(obj->isCurrent());
    }

}

void PolyAreaGraphicsItem::updatePolygon()
{
    setPolygon(keyframe->polygon);

    foreach(PolyPointGraphicsItem * item, polyPoints) {
        scene()->removeItem(item);
    }
    qDeleteAll(polyPoints);
    polyPoints.clear();

    foreach(QPointF point, keyframe->polygon){
        PolyPointGraphicsItem* pointItem = new PolyPointGraphicsItem(scene(), point);
        polyPoints.append(pointItem);
        connect(pointItem, SIGNAL(moving()), this, SLOT(redrawFromPolyPoints()));
        connect(pointItem, SIGNAL(moved()), this, SLOT(polyPointsChanged()));
    }
}

void PolyAreaGraphicsItem::filterKeyFrameChange(int f)
{
    if(f & (KeyFrame::PolygonChange)) {
        updatePolygon();
    }
}

void PolyAreaGraphicsItem::filterStateChange(int f)
{
    if(f & DataState::CurrentObjectChange) {
        updateStyle();
    }
}

void PolyAreaGraphicsItem::movePolygon()
{
    Eigen::Affine2d t = Eigen::Translation2d(move.x(), move.y()) * Eigen::Affine2d::Identity();
    keyframe->transformPolygon(t);

    if(keyframe->type == KeyFrame::Automatic) keyframe->setType(KeyFrame::SetVisible);
}

void PolyAreaGraphicsItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    QGraphicsPolygonItem::mousePressEvent(event);
    move = QPointF(0.0,0.0);

    QPointF mouseMoveStart = event->scenePos();
    QRectF imageRect = g_mainWindow->frameViewer->pixmapItem->sceneBoundingRect();
    QPointF topLeftBoundary = mouseMoveStart - polygon().boundingRect().topLeft();
    QPointF bottomRightBoundary = imageRect.bottomRight() + (mouseMoveStart - polygon().boundingRect().bottomRight());
    moveBoundingBox = QRectF(topLeftBoundary, bottomRightBoundary);
}

void PolyAreaGraphicsItem::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    if(keyframe->object->isCurrent()) {
        QPointF mousePos = event->scenePos();

        QPointF threshMousePos = mousePos;
        threshMousePos.setX(qMin(moveBoundingBox.right(), qMax(moveBoundingBox.left(), threshMousePos.x())));
        threshMousePos.setY(qMin(moveBoundingBox.bottom(), qMax(moveBoundingBox.top(), threshMousePos.y())));

        QPointF threshLastMousePos = event->lastScenePos();
        threshLastMousePos.setX(qMin(moveBoundingBox.right(), qMax(moveBoundingBox.left(), threshLastMousePos.x())));
        threshLastMousePos.setY(qMin(moveBoundingBox.bottom(), qMax(moveBoundingBox.top(), threshLastMousePos.y())));

        QPointF diff = threshMousePos - threshLastMousePos;
        move += diff;

        foreach(PolyPointGraphicsItem* polyPoint, polyPoints) {
            polyPoint->moveBy(diff.x(), diff.y());
        }

        setPos(pos()+diff);
    }
}

void PolyAreaGraphicsItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    QGraphicsPolygonItem::mouseReleaseEvent(event);
    if(move != QPointF(0.0,0.0)) {
        // object moved
        if(keyframe->object->isCurrent()) movePolygon();
        move = QPointF(0.0,0.0);
        setPos(0,0);
    } else {
        // object clicked
        if(event->modifiers() & Qt::ControlModifier) g_state->setCurrentObject(keyframe->object->id);
    }
}
