#include "trackedobject.h"

#include "../globalobjects.h"

int TrackedObject::newID = 0;

TrackedObject::TrackedObject(QObject *parent) :
    QObject(parent),
    id(newID++)
{
    name = "new object "+QString::number(id+1);
    for(int frame = 0; frame < g_state->vid->frameCount; frame++) {
        KeyFrame* kf = new KeyFrame(this, frame, KeyFrame::NotSet);
        connect(kf, SIGNAL(hasChanged(int)), this, SLOT(keyFrameChanged(int)));
        keyFrames.append(kf);
    }

    FrameSpan* span = new FrameSpan(this, FrameSpan::NotSet, 0, g_state->vid->frameCount -1);
//    connect(span, SIGNAL(hasChanged(int)), this, SLOT(frameSpanChanged(int)));
    for(int frame = 0; frame < g_state->vid->frameCount; frame++) {
        frameSpans.append(span);
    }

    color = g_state->colors[this->id % g_state->colors.count()];
}

void TrackedObject::changeName(QString txt)
{
    name = txt;
    emit hasChanged(NameChange);
}

KeyFrame *TrackedObject::getCurrentKeyFrame()
{
    return keyFrames[g_state->currentFrame];
}

QList<FrameSpan *> TrackedObject::getFrameSpans()
{
    QList<FrameSpan *> frameSpansReturn;
    frameSpansReturn.append(frameSpans.at(0));
    for(int i=frameSpans.at(0)->nextFrame;i<g_state->vid->frameCount-1;i++){
        frameSpansReturn.append(frameSpans.at(i));
        i=frameSpans.at(i)->nextFrame;
    }
    return frameSpansReturn;
}

TrackedObject::TrackedObjectLess TrackedObject::combareByZOrderOfFrame(int frame)
{
    return TrackedObjectLess(frame);
}

int TrackedObject::getZOrderRank(int frame)
{
    QList<TrackedObject *> l = g_state->getZOrderSortedList(frame);
    for(int i=0; i<l.count(); ++i) {
        if(l[i]->id == this->id) return i;
    }
    return 0;
}


void TrackedObject::keyFrameChanged(int f)
{
    // ----------------------------------------------------------------------
    // If keyframe has been set/unset, framespans need to be split or merged
    // ----------------------------------------------------------------------

    // get signal sending keyframe
    KeyFrame *kf = dynamic_cast<KeyFrame*>(sender());
    int frame = kf->frame;

    // only consider type changes
    if(!(f & KeyFrame::TypeChange)) return;
    // do nothing if typechange is automatic (= from morphing)
    if(kf->type == KeyFrame::Automatic) return;


    // specific handling for boundary frames
    bool isBoundaryFrame = (frame == 0 || frame == g_state->vid->frameCount-1);
    // keyframe was set before, if there are different spans surrounding
    // cannot be retrieved for boundary frames
    bool kfWasSetBefore = !isBoundaryFrame && (frameSpans[frame-1] != frameSpans[frame]);
    bool kfIsNowSet = (kf->type == KeyFrame::SetVisible || kf->type == KeyFrame::SetOccluded);

    // CASE 1: keyframe has been unset and was set before
    if(!kfIsNowSet && (kfWasSetBefore || isBoundaryFrame)) {

        // part 1: deleting (not necessary for boundaries)
        if(!isBoundaryFrame) {

            // keep left span
            FrameSpan *spanToKeep = frameSpans[frame-1];

            // delete right span
            FrameSpan *spanToDelete = frameSpans[frame];

            // adjust endpoint
            spanToKeep->nextFrame = spanToDelete->nextFrame;

            // adjust pointers in framespans array
            for(int i = frame; i < spanToDelete->nextFrame; i++) frameSpans[i] = spanToKeep;

            // set type to notype if both have different types
            if(spanToDelete->type != spanToKeep->type) {
                spanToKeep->setType(FrameSpan::NotSet);
            }

            // delete
            spanToDelete->deleteLater();
        }

        // part 2: adjust span type
        // idea: set type to notype if there are no keyframes left in the other direction

        // to the left
        if(frame == 0) frameSpans[frame]->setType(FrameSpan::NotSet);
        else {
            KeyFrame *prevKf = keyFrames[frameSpans[frame]->prevFrame];
            if(prevKf->type == KeyFrame::NotSet) {
                frameSpans[frame]->setType(FrameSpan::NotSet);
            }
        }
        // to the right
        if(frame == g_state->vid->frameCount-1) frameSpans[frame]->setType(FrameSpan::NotSet);
        else {
            KeyFrame *nextKf = keyFrames[frameSpans[frame]->nextFrame];
            if(nextKf->type == KeyFrame::NotSet) {
                frameSpans[frame]->setType(FrameSpan::NotSet);
            }
        }

        // part 3: clear automatic keyframes over affected area
        for(int i = frameSpans[frame]->prevFrame+1; i < frameSpans[frame]->nextFrame; i++) {
            keyFrames[i]->setNewPolygon(QPolygonF());
            keyFrames[i]->setType(KeyFrame::NotSet);
        }

        emit hasChanged(FrameSpanChange);

    // CASE 2: kf has been set and wasnt set before
    } else if(kfIsNowSet && (!kfWasSetBefore || isBoundaryFrame)){

        // part 1: splitting (not necessary for boundaries)
        if(!isBoundaryFrame) {

            // old span has to be split in two
            FrameSpan *oldSpan = frameSpans[frame];

            // create new span (same type)
            FrameSpan *newSpan = new FrameSpan(this, oldSpan->type, frame, oldSpan->nextFrame);

            // adjust endpoint of previous span
            oldSpan->nextFrame = frame;

            // adjust pointers in framespans array
            for(int i = frame; i < newSpan->nextFrame; i++) frameSpans[i] = newSpan;
        }

        // part 2: adjust span type
        // idea: set visible, if not set before and only if there is a keyframe in that direction

        // to the left
        if(frame != 0 && frameSpans[frame-1]->type == FrameSpan::NotSet) {
            KeyFrame *prevKf = keyFrames[frameSpans[frame-1]->prevFrame];
            if(prevKf->type == KeyFrame::SetOccluded || prevKf->type == KeyFrame::SetVisible) {
                frameSpans[frame-1]->setType(FrameSpan::Visible);
            }
        }
        // to the right
        if(frame != g_state->vid->frameCount-1 && frameSpans[frame]->type == FrameSpan::NotSet) {
            KeyFrame *nextKf = keyFrames[frameSpans[frame]->nextFrame];
            if(nextKf->type == KeyFrame::SetOccluded || nextKf->type == KeyFrame::SetVisible) {
                frameSpans[frame]->setType(FrameSpan::Visible);
            }
        }

        // part 3: clear automatic keyframes over affected areas
        if(frame != 0) {
            for(int i = frameSpans[frame-1]->prevFrame+1; i < frame; i++) {
                keyFrames[i]->setNewPolygon(QPolygonF());
                keyFrames[i]->setType(KeyFrame::NotSet);
            }
        }
        if(frame != g_state->vid->frameCount-1) {
            for(int i = frame+1; i < frameSpans[frame]->nextFrame; i++) {
                keyFrames[i]->setNewPolygon(QPolygonF());
                keyFrames[i]->setType(KeyFrame::NotSet);
            }
        }

       emit hasChanged(FrameSpanChange);
    }
    return;
}


bool TrackedObject::isCurrent()
{
    return (this->id == g_state->currentObjectID);
}
